#include <stdio.h>
#include <math.h>
#include <avs/flow.h> 
#include <avs/avs_data.h>
#include <avs/field.h>
#include <time.h>

#define PERFRAME 30
extern int binomc_();
extern int BINOMC_();

AVSinit_modules()
{
  int fin();

  AVSmodule_from_desc(fin);
}

/*  The routine "fin" is the description routine.  */
fin()
{
  int fin_compute();	 /* declare the compute function (below) */
  int in_port,out_port;	 /* temporary to hold the port number */
  int parm;		 /* temporary to hold the parm number */

  /* Set the module name and type */
  AVSset_module_name("AMC_SUN", MODULE_MAPPER);
  
  /* Create an input port for the required field input */
  in_port = AVScreate_input_port("Input Field",
				 "field 1D uniform scalar float", REQUIRED);  
  /* Create output ports for the result */
  out_port = AVScreate_output_port("Field Output", 
				   "field 1D scalar uniform float");
  out_port = AVScreate_output_port("AMC_bias", "real"); 

  /* Add parameter:  */
  parm = AVSadd_float_parameter("V_AMC", 1.0, 0.0, 5.0);
  parm = AVSadd_parameter("points/frame","integer", PERFRAME, 1, 100);  

  /* Tell avs what subroutine to call to do the compute */
  AVSset_compute_proc(fin_compute);
}

/*
 * fin_compute is the compute routine.  It is called whenever AVS wants to 
 * compute new model results.  The arguments are: the value of the input
 * field, the new output field (doubly indirected), the parameter 
 * values in the order same as declaration above.
 * Note the order is always inputs, outputs, parameters. 
 */
static float v_amc;
static int NREPB=100;
static float sz,k,rate,t,tdiv,div,call;

fin_compute(input, output, pbias, pv_amc, size)
AVSfield_float *input, **output;
float *pv_amc, **pbias;
int size;
{
  register int i;  
  AVSfield_float *temp_field = NULL;
  static float bias;   
  int dims[1];
   
  float camc;
  float *dataptr = input->data;
  float implied_vol(),tmp;

  sz = *(dataptr++);  k = *(dataptr++); rate = *(dataptr++);
  t = *(dataptr++); tdiv = *(dataptr++); div = *(dataptr++);
  call = *dataptr;

  /* Allocate space for output */
  dims[0] = size;
  /*fprintf(stderr,"allocate memory ...\n");*/
  temp_field = (AVSfield_float *)AVSdata_alloc
    ("field 1D scalar uniform float",dims);
  if (temp_field == NULL) {
    AVSerror("Allocation of output field failed.");
    return(0);
  }
     
  if (*output) {
    for (i=0; i<(size-1); i++) 
      temp_field->data[i] = (*output)->data[i+1];
    /* fprintf(stderr,"deallocate memory ...\n");*/
    AVSdata_free("field",*output);
  }
  
  if (sz<0.0) {  /* use own optimizing function */
    sz = -sz;
    if ((tmp = implied_vol()) > 0.0) 
      v_amc = tmp;
  }
  else {
    if (*pv_amc != 0.0) v_amc = *pv_amc;
  }
  
  AVSmodify_float_parameter("V_AMC",AVS_VALUE, v_amc, NULL, NULL);
 
  binomc_ (&sz,&k,&rate,&t,&v_amc,&tdiv,&div,&NREPB,&camc);
  temp_field->data[size-1] = bias =  camc - call ;

  fprintf(stderr," AMC bias=%8.5f",bias);
  *output = temp_field;
  *pbias = &bias; 
 
  /* When we're done, we return 1 to indicate success */
  return(1);
}

#define nmax 24

float implied_vol()
/* Use bisection method to find volatility */
{
  float v1,v2,v0,vmid,fmid,dv,f;
  int i;

  v1 = 0.0000006; v2 = 1.2;
  binomc_ (&sz,&k,&rate,&t,&v2,&tdiv,&div,&NREPB,&fmid);
  fmid -= call;
  binomc_ (&sz,&k,&rate,&t,&v1,&tdiv,&div,&NREPB,&f);
  f -= call;
  if (f*fmid >= 0.0) {
    fprintf(stderr,"implied_vol: root not bracketed\n");
    return(-1.0);
  }

  if (f < 0.0) {
    v0 = v1;
    dv = v2 - v1;
  }
  else {
    v0 = v2;
    dv = v1 - v2;
  }
  for (i = 0; i < nmax; i++) {
    dv = dv/2.0;
    vmid = v0 + dv;
    binomc_ (&sz,&k,&rate,&t,&vmid,&tdiv,&div,&NREPB,&fmid);
    fmid -= call;
    if (fmid < 0.0) v0 = vmid;
  }

  return(v0);
}
